"""
python v3.9.0
@Project: hotpot
@File   : cc
@Auther : Zhiyuan Zhang
@Data   : 2023/6/4
@Time   : 21:05

This module is to perform `Task` about Coordination Chemistry (cc)
"""
from ..cheminfo import Molecule

class Ligand(Molecule):
    """ Represent a ligand """


class MetalLigandPair(Molecule):
    """ Represent a metal-organic pair """

